/************************************************************************

    JOYCE v2.1.6 - Amstrad PCW emulator

    Copyright (C) 1996, 2001, 2004  John Elliott <jce@seasip.demon.co.uk>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*************************************************************************/

#include "Pcw.hxx"

#if 0	/* This file is a failed experiment */

#include "bw.xbm"
#include "hatches.xbm"
#include "halftone.xbm"

#define nPTSIN  1	/* Indices into the control array for sizes of */
#define nPTSOUT 2	/* the other arrays */
#define nINTIN  3
#define nINTOUT 4
#define nESC    5

#define LINE_COLOUR   ((line_colour   & gsxfg_and) | gsxfg_or)
#define MARKER_COLOUR ((marker_colour & gsxfg_and) | gsxfg_or)
#define FILL_COLOUR   ((fill_colour   & gsxfg_and) | gsxfg_or)
#define TEXT_COLOUR   ((text_colour   & gsxfg_and) | gsxfg_or)
#define X0 points_in[0][0]
#define Y0 points_in[0][1]
#define X1 points_in[1][0]
#define Y1 points_in[1][1]
#define X2 points_in[2][0]
#define Y2 points_in[2][1]
#define X3 points_in[3][0]
#define Y3 points_in[3][1]

PcwGSX::PcwGSX(PcwSystem *s) : PcwSdlTerm("gsx", s)
{
	m_sys = s;
	oldtype = -1;
	text_sin = 0;
	text_cos = 100;
	line_colour = line_style = marker_colour = marker_style =
	        text_style = text_colour = fill_colour = fill_style = 
		fill_index = locator_mode = 0;
	line_size = marker_size = text_size = 0;
	initPalette();
	m_term = NULL;
}

PcwGSX::~PcwGSX()
{
	if (m_term) delete m_term;
	m_term = NULL;
}

static word dflt_col[] = 
{
	0, 0, 0,     1000, 0, 0,     0, 1000, 0,    1000, 1000, 0,
	0, 0, 1000,  1000, 0, 1000,  0, 1000, 1000
};



/********************************************************************/

/* Colour handling in JOYCE GSX */
/********************************/
/* JOYCE has 64 fixed colours and 192 user-defined ones. The first 
 * 64 user-defined colours cannot be changed once defined, while the
 * last 128 are automatically reused when the 257th colour is requested.
 *
 * JOYCE GSX allows 128 colours to be used. These are initialised to use
 * the 64 standard colours and any 64 user-defined colours.
 *
 * When GSX requests a palette change, a check is made on the JOYCE number
 * of the colour. If it is <128, then setInk() is used to allocate a new
 * colour. If it is >=128, the JOYCE colour is redefined.
 *
 * To muddy the issue further, colours are actually applied using macros
 * (eg LINE_COLOUR) so that changes made by the GSX writing mode take effect.
 */



void PcwGSX::initPalette(void)
{
	int n;

	if (!m_sysVideo) return;

	memset(request_col, 0, sizeof(request_col));
	memcpy(request_col, dflt_col, sizeof(dflt_col));
	m_colourMap[0] = m_sysVideo->colourToIndex(0x00, 0x00, 0x00);
	m_colourMap[1] = m_sysVideo->colourToIndex(0xFF, 0x00, 0x00);
	m_colourMap[2] = m_sysVideo->colourToIndex(0x00, 0xFF, 0x00);
	m_colourMap[3] = m_sysVideo->colourToIndex(0xFF, 0xFF, 0x00);
	m_colourMap[4] = m_sysVideo->colourToIndex(0x00, 0x00, 0xFF);
	m_colourMap[5] = m_sysVideo->colourToIndex(0xFF, 0x00, 0xFF);
	m_colourMap[6] = m_sysVideo->colourToIndex(0x00, 0xFF, 0xFF);

	for (n = 7; n < 128; n++)
	{
		m_colourMap[n] = m_sysVideo->colourToIndex(0xFF, 0xFF, 0xFF);
		request_col[n][0] = 1000;
		request_col[n][1] = 1000;
		request_col[n][2] = 1000;
	}
}



void PcwGSX::letterPixel(int xb, int yb, int x, int y, Uint32 tcol)
{
	int x1, y1;

	x1 = (x * text_cos - y * text_sin) / 100;
	y1 = (x * text_sin + y * text_cos) / 100;

	pixelColor(m_surface, xb + x1, yb - y1, tcol);
}
#if 0
static GrLineOption GSXopt;
static GrPattern GSXfill;

static byte pat2[] = {7, 9};
static byte pat3[] = {3, 5};
static byte pat4[] = {5, 3, 3, 5};
static byte pat5[] = {12, 4};

static void SetGSXopt(void)
{
	GSXopt.lno_color = LINE_COLOUR;
	GSXopt.lno_width = line_size;
	
	switch(line_style)
	{
		default:
		case 1:	GSXopt.lno_pattlen = 0;
			break;

		case 2: GSXopt.lno_pattlen = 2;
			GSXopt.lno_dashpat = pat2;
			break;

		case 3: GSXopt.lno_pattlen = 2;
			GSXopt.lno_dashpat = pat3;
			break;

		case 4: GSXopt.lno_pattlen = 4;
			GSXopt.lno_dashpat = pat4;
			break;

		case 5: GSXopt.lno_pattlen = 2;
			GSXopt.lno_dashpat = pat5;
			break;
	}
}


static void SetGSXfill(void)
{
	int fill_offs = 8*(fill_index - 1);
	GrBitmap *pbm = (GrBitmap *)&GSXfill;

	pbm->bmp_ispixmap = 0;
	pbm->bmp_height   = 8;
	pbm->bmp_fgcolor  = FILL_COLOUR;
	pbm->bmp_bgcolor  = gsxbg;
	pbm->bmp_memflags = 0;

	switch(fill_style)
	{
		case 0:	pbm->bmp_fgcolor = GrNOCOLOR;
			pbm->bmp_bgcolor = GrNOCOLOR;
			pbm->bmp_data    = (char *)bw_bits;
			break;
		case 1:	pbm->bmp_fgcolor = FILL_COLOUR;
			pbm->bmp_bgcolor = gsxbg;
			pbm->bmp_data    = (char *)(bw_bits + 8);
			break;
		case 2:	if (fill_offs >= halftone_height) fill_offs = 0;
			pbm->bmp_fgcolor = FILL_COLOUR;
			pbm->bmp_bgcolor = gsxbg;
			pbm->bmp_data    = (char *)(halftone_bits + fill_offs);
			break;
		case 3:	if (fill_offs >= hatches_height) fill_offs = 0;
			pbm->bmp_fgcolor = FILL_COLOUR;
			pbm->bmp_bgcolor = gsxbg;
			pbm->bmp_data    = (char *)(hatches_bits + fill_offs);
			break;
	}
}

#endif


void PcwGSX::getPtsin()
{
	int n;
	int max = getArray(ctrl, nPTSIN); 
	
	if (max > 75) max = 75;

	for (n = 0; n < max; n++)
	{
		points_in[n][0] = getArray(ptsin, n * 2);
		points_in[n][1] = 599 - getArray(ptsin, n * 2 + 1);
	}
}


void PcwGSX::gsxPolyLine(int npts, Uint32 lineCol)
{
	int n;
	for (n = 0; n < (npts-1); n++)
	{
		gsxLine(points_in[n][0],   points_in[n][1],
			points_in[n+1][0], points_in[n+1][1], lineCol);
	}
}


void PcwGSX::gsxLine(int x1, int y1, int x2, int y2, Uint32 colour)
{
	aalineColor(m_surface, x1, y1, x2, y2, colour);
}

#if 0
static void GSXCircleArc(int x0, int y0, int x3, int f1, int f2, int f3, int line_col)
{
	GSXopt.lno_color = line_col;
	GrCustomCircleArc(x0, y0, x3, f1, f2, f3, &GSXopt);
/*	GrCircleArc(x0, y0, x3, f1, f2, f3, line_col); */
}

static void GSXCircle(int x, int y, int r, int col)
{
	GSXopt.lno_color = col;
	GrCustomCircle(x, y, r, &GSXopt);
/*	GrCircle(x, y, r, col); */
}

static void GSXBox(int x1, int y1, int x2, int y2, int line_col)
{
	GSXopt.lno_color = line_col;
	GrCustomBox(x1, y1, x2, y2, &GSXopt);
/*	GrBox(x1, y1, x2, y2, line_col); */
}



static void GSXPolygon(int npts, int points[][], int line_col)
{
	GSXopt.lno_color = line_col;
	GrCustomPolygon(npts, points, &GSXopt);
/*	GrPolygon(npts, points, line_col); */
}



static void GSXmarker(int x,int y)
{
	int lsize = (marker_size * 2) / 3;
	switch(marker_style)
	{
		case 0:
		case 1:		GrPlot(x, y, MARKER_COLOUR);
				break;
	
		case 2:		GrPlot(x, y, MARKER_COLOUR);
				GrHLine(x - marker_size, x + marker_size, y, MARKER_COLOUR);
				GrVLine(x, y - marker_size, y + marker_size, MARKER_COLOUR);
				break;
		default:
		case 3:		GrPlot(x, y, MARKER_COLOUR);
				GrLine(x - lsize, y - lsize, x + lsize, y + lsize, MARKER_COLOUR);
				GrLine(x + lsize, y - lsize, x - lsize, y + lsize, MARKER_COLOUR);
				GrVLine(x, y - marker_size, y + marker_size, MARKER_COLOUR);
				GrHLine(x - marker_size, x + marker_size, y, MARKER_COLOUR);
				break;

		case 4:	 	GrCircle(x, y, marker_size, MARKER_COLOUR);
				break;

		case 5:		GrPlot(x, y, MARKER_COLOUR);
				GrLine(x - marker_size, y - marker_size, x + marker_size, y + marker_size, MARKER_COLOUR);
				GrLine(x + marker_size, y - marker_size, x - marker_size, y + marker_size, MARKER_COLOUR);
				break;
	}

}

static char GSXcur[]=
{
	0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,1,3,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,1,3,1,0,0,0,0,0,0,0, 
	0,0,0,0,0,0,1,4,1,0,0,0,0,0,0,0,
	1,1,1,1,1,1,1,4,1,1,1,1,1,1,1,0,
	1,2,2,3,3,4,4,5,4,4,3,3,2,2,1,0,
	1,1,1,1,1,1,1,4,1,1,1,1,1,1,1,0, 
	0,0,0,0,0,0,1,4,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,1,3,1,0,0,0,0,0,0,0, 
	0,0,0,0,0,0,1,3,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,1,2,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,1,1,1,0,0,0,0,0,0,0,
	0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0
};


static long GSXcurtbl[6];
static GrCursor *GSXcursor;

static void MakeCursor(void)
{
	GSXcurtbl[0] = GrBlack();
	GSXcurtbl[1] = GrWhite();
	setInk((int *)&GSXcurtbl[2], 0x80, 0x80, 0xC0);
	setInk((int *)&GSXcurtbl[3], 0x40, 0x40, 0xC0);
	setInk((int *)&GSXcurtbl[4], 0x40, 0x40, 0x80);
	setInk((int *)&GSXcurtbl[5], 0x00, 0x00, 0x80);

	GSXcursor = GrBuildCursor(GSXcur, 1, 16, 16, 7, 7, GSXcurtbl);
}


#endif
void PcwGSX::getColour(Uint32 *col, word array, word off)
{
	Uint8 r, g, b;
	word nc = getArray(array, off);
	if (nc >= 128) nc = 127;

	m_sysVideo->indexToColour(nc, &r, &g, &b);

	*col = r;
	*col = (*col << 8) | g;
	*col = (*col << 8) | b;
	*col = (*col << 8) | 0xff;	/* Alpha */
}
#if 0

static void vgastr(char *s)
{
	while (*s)
	{
		vga_putchar(*s);
		++s;
	}
}
#define ESC "\0x1B"

#endif
void PcwGSX::setColour(Uint32 *ptr)
{
	int col = getWord(intin);
	getColour(ptr, intin, 0);
	m_sys->m_cpu->storeW(intout, (col > 127) ? 127 : col);
	putArray(ctrl, nINTOUT, 1);	
}
#if 0

void SetSize(int *ptr, int which)
{
	*ptr = getArray(ptsin, which);
}


#define SetTextHeight   SetSize(&text_size, 1);
#define SetLineWidth	SetSize(&line_size, 0);
#define SetMarkerHeight SetSize(&marker_size, 1);
#endif

#define setLineColour   setColour(&line_colour)
#define setMarkerColour setColour(&marker_colour)
#define setTextColour   setColour(&text_colour)
#define setFillColour   setColour(&fill_colour)




/********************************************************************/

static const signed short devdata[45] =
{
	799,	/* Width */
	599,	/* Height */
	1,	/* CRT */
	350,	/* Pixel width */
	350,	/* Pixel height */
	1,	/*  1 character size */
	5,	/*  5 line styles */
	12,	/* 12 line widths */
	5,	/*  5 marker styles */
	12,	/* 12 marker sizes */
	1,	/* One font */
	1,	/* One halftone (actual number set at runtime) */
	1,	/* One hatch    (actual number set at runtime) */
	128,	/* Max displayable colours */
	4,	/* Four GDPs */
	  1,	/*   GDP 1 = bar */
 	  2,	/*   GDP 2 = arc */
	  3,	/*   GDP 3 = pie slice */
	  4,	/*   GDP 4 = circle */
	  -1,	/*   GDP 5 = end of list */
	  0,	/*   GDP 6 = unused */
	  0,	/*   GDP 7 = unused */
	  0,	/*   GDP 8 = unused */
	  0,	/*   GDP 9 = unused */
	  0,	/*   GDP 10 = unused */
	  3,	/*   GDP 1 = filled bar */
	  0,	/*   GDP 2 = arc */
	  3,	/*   GDP 3 = filled pie slice */
	  3,	/*   GDP 4 = filled circle */
	  -1,	/*   GDP 5 = end of list */
	  0,	/*   GDP 6 = unused */
	  0,	/*   GDP 7 = unused */
	  0,	/*   GDP 8 = unused */
	  0,	/*   GDP 9 = unused */
	  0,	/*   GDP 10 = unused */
	1,	/* Colour */
	1,	/* Text rotation available */
	1,	/* Area fill available */
	0,	/* Read cell array not available */
	0,	/* More than 32767 colours */
	0,	/* No locator */
	0,	/* No valuator */
	0,	/* No choice */
	0,	/* Output only */
};

static const word devscales[12] =
{
	0, 16,	/* Min character height */
	0, 16,	/* Max character height */
	1, 0,	/* Min line width */
	12, 0,	/* Max line width */
	0, 1,	/* Min marker height */
	0, 12, 	/* Max marker height */
};


void PcwGSX::openWorkstation(void)
{
	int n;
	int maxb = (halftone_height > hatches_height ? halftone_height : hatches_height);

        m_oldTerm = m_sys->m_activeTerminal;
	m_sys->selectTerminal(this);

	m_term = new SdlTerm(m_sysVideo->getSurface());
	SDL_Rect &rcTerm = m_term->getBounds();

	initPalette();
//	makeCursor();

	gsxbg = 0;
	gsxfg_and = (~0);		/* GSX foreground modifiers */
	gsxfg_or  = 0;
	text_sin = 0;
	text_cos = 100;

	line_style    = getArray(intin, 1); 
	if (line_style < 1 || line_style > 5) line_style = 1;
	getColour(&line_colour, intin, 2);
	marker_style  = getArray(intin, 3);
	if (marker_style > 5 || marker_style < 1) marker_style = 3;
	getColour(&marker_colour, intin, 4);
	text_style    = 1; /* getArray(intin, 5); */
	getColour(&text_colour, intin, 6);
	fill_style    = getArray(intin, 7);
	if (fill_style > 3) fill_style = 2;
	fill_index    = getArray(intin, 8);
	if (fill_index > (maxb / 8 + 1) || fill_index < 1) fill_index = 1;
	getColour(&fill_colour, intin, 9);

	line_size   = 1;
	marker_size = 1;
	text_size   = 16;

	putArray(ctrl, nPTSOUT, 6);
	putArray(ctrl, nINTOUT, 45);

	putArray(intout, 0, rcTerm.w);
	putArray(intout, 1, rcTerm.h);

	for (n = 2; n < 45; n++) putArray(intout, n, devdata[n]);
	for (n = 0; n < 12; n++) putArray(ptsout, n, devscales[n]);

	n = hatches_height / 8;  putArray(intout, 12, n);
	n = halftone_height / 8; putArray(intout, 11, n);
	
	locator_mode = 1;
	/* There's always a mouse if (GrMouseDetect())  */
	putArray(intout, 41, 1);

//	GrSetClipBox(0, 0, 800, 600);
//	vga_cursor(0);
	clearPicture();
}


void PcwGSX::closeWorkstation()
{
	m_sys->selectTerminal(m_oldTerm);
	delete m_term;
	m_term = NULL;
}


void PcwGSX::clearPicture()
{
	m_term->setBackground(0, 0, 0);
	m_term->clearWindow();
}

void PcwGSX::updateWorkstation()
{
	SDL_UpdateRect(m_surface, 0, 0, m_surface->w, m_surface->h);
}

#if 0


static void Escape(void)
{
	word txtc;

	switch(getArray(ctrl, nESC))
	{
		case 1:	
		putArray(ctrl, nINTOUT, 2);
		putArray(intout, 0, 37);
		putArray(intout, 1, 100);
		break;

		case 2: /* Exit gfx mode */
		GrClearScreen(dblk);
		vga_cursor(1);
		break;

		case 3: /* Enter gfx mode */
		vga_cursor(0);
		GrClearScreen(GrBlack());
		break;

		case 4: /* Cursor up */ vgastr(ESC "A"); break;
		case 5: /* Cursor dn */ vgastr(ESC "B"); break;
		case 6: /* Cursor lt */ vgastr(ESC "C"); break;
		case 7: /* Cursor rt */ vgastr(ESC "D"); break;
		case 8: /* Cursor hm */ vgastr(ESC "H"); break;
		case 9: /* Erase EOS */ vgastr(ESC "J"); break;
		case 10: /* Erase EOL */ vgastr(ESC "K"); break;

		case 11: /* Position cursor */
		vgastr(ESC "Y");
		vga_putchar(getArray(intin, 0) + 32);
		vga_putchar(getArray(intin, 1) + 32);
		break;

		case 12: /* Text out */
		for (txtc = 0; txtc < getArray(ctrl, nINTIN); txtc++)
		{
			vga_putchar(getArray(intin, txtc));
		}
		break;

		case 13: /* Rev vid   */ vgastr(ESC "p"); break;
		case 14: /* True vid  */ vgastr(ESC "q"); break;
		
		case 15:
		putArray(ctrl, nINTOUT, 2);
		putArray(intout, 0, vga_y);
		putArray(intout, 1, vga_x);
		break;

		case 16:		/* No tablet */
		putArray(ctrl, nINTOUT, 1);
		putArray(intout, 0, 0);
		break;
		
		case 17:		/* No hardcopy */
		break;

		case 18:
		getptsin();
		GrDisplayCursor(GSXcursor);
		GrMoveCursor(GSXcursor, X0, Y0);
		break;

		case 19:
		GrEraseCursor(GSXcursor);
		break;
	}

}


#endif
void PcwGSX::drawPolyline(void)
{
	getPtsin();	
	gsxPolyLine(getArray(ctrl, nPTSIN), LINE_COLOUR);
}

#if 0
static void DrawPolymarker(void)
{
	int max = getArray(ctrl, nPTSIN);
	int n;

	getptsin();
	for (n = 0; n < max; n++)
	{
		GSXmarker(points_in[n][0], points_in[n][1]);
	}
}


static void DrawText(void)
{
	int l = getArray(ctrl, nINTIN);
	int p;
	byte *pFont;
	word x, y;
	word xb = getArray(ptsin, 0);
	word yb = 599 - getArray(ptsin, 1);
	byte mask;
	int col = TEXT_COLOUR;

	for (p = 0; p < l; p++)
	{
		pFont = gsxfont + 16 * getArray(intin, p);		
		for (y = 0; y < 16; y++)
		{
			mask = 0x80;
			for (x = 0; x < 8; x++)
			{
				GSXLetterPixel(xb, yb, x + 8*p, 15 - y,
					      ((pFont[y] & mask) ? col : gsxbg));
				mask = mask >> 1;
			}
		}
	}
}	



static void DrawPolygon(void)
{
	getptsin();	
	if (fill_style) GrPatternFilledPolygon(getArray(ctrl, nPTSIN), points_in, &GSXfill); 
	GSXPolygon(getArray(ctrl, nPTSIN), points_in, FILL_COLOUR);
}


static void OutputIndexArray(void)
{
	int x , y, rx, ry, c;
	int rowlen  = getArray(ctrl, 5);
	int rowused = getArray(ctrl, 6);
	int nrows   = getArray(ctrl, 7);
	int wmode   = getArray(ctrl, 8);
	
	switch(wmode)
	{
		case 0: wmode = 0;     break;
		case 1: wmode = GrOR;  break;
		case 2: wmode = GrXOR; break;
		case 3: wmode = GrAND; break;
	}

	getptsin();
	ry = 0;
	for (y = Y0; y < Y1; y++)
	{
		rx  = 0;
		for (x = X0; x < X1; x++)
		{
			GetColour(&c, intin, ry * rowlen + rx);
			if (wmode == GrAND) c = gsxbg; else c |= wmode;
			GrPlot(x, y, c);
			++rx; if (rx >= rowused) rx = 0;
		}
		++ry; if (ry >= nrows) ry = 0;
	}
}






static void GDP(void)
{
	getptsin();
	switch(getArray(ctrl,nESC))
	{
		case 1: /* bar */
		if (fill_style) GrPatternFilledBox(X0, Y0, X1, Y1, &GSXfill);
		GSXBox                            (X0, Y0, X1, Y1, FILL_COLOUR);
		break;

		case 2: /* arc */
		GSXCircleArc(X0, Y0, X3, getArray(intin, 0), getArray(intin,1), GR_ARC_STYLE_CLOSE1, FILL_COLOUR);
		break;

		case 3: /* pie slice */

		if (fill_style) GrPatternFilledCircleArc(X0, Y0, X3, getArray(intin, 0), getArray(intin,1), GR_ARC_STYLE_CLOSE1, &GSXfill);
		GSXCircleArc(X0, Y0, X3, getArray(intin, 0), getArray(intin,1), GR_ARC_STYLE_CLOSE1, FILL_COLOUR);
		GSXLine(X0, Y0, X1, Y1, FILL_COLOUR);
		GSXLine(X0, Y0, X2, Y2, FILL_COLOUR);
		break;

		case 4: /* circle */
		if (fill_style) GrPatternFilledCircle(X0, Y0, X2, &GSXfill); 
		GSXCircle                     (X0, Y0, X2, FILL_COLOUR); 
		break;
	}
}


static void SetTextDirection(void)
{
	text_sin = (short)getArray(intin, 2); /* they're _signed_ integers */
	text_cos = (short)getArray(intin, 1);
}

#endif
void PcwGSX::setColourIndex(void)
{
	int red   = getArray(intin, 1) * 255 / 1000;	
	int green = getArray(intin, 2) * 255 / 1000;	
	int blue  = getArray(intin, 3) * 255 / 1000;	
	int ncol  = getArray(intin, 0);

	if (ncol > 127) ncol = 127;

	request_col[ncol][0] = getArray(intin, 1);
	request_col[ncol][1] = getArray(intin, 2);
	request_col[ncol][2] = getArray(intin, 3);

	m_colourMap[ncol] = m_sysVideo->colourToIndex(red, green, blue);
}

void PcwGSX::setLineStyle(void)
{
	word style = getWord(intin);
	if (style > 5 || style < 1) style = 1;
	m_sys->m_cpu->storeW(intout, style);
	putArray(ctrl, nINTOUT, 1);
	line_style = style;
}

/*
static void SetMarkerStyle(void)
{
	word style = getWord(intin);
	if (style > 5 || style < 1) style = 3;
	PutWord(intout, style);
	putArray(ctrl, nINTOUT, 1);
	marker_style = style;
}


static void SetTextStyle(void)
{
	word style = 1;	// Only one text style //
	PutWord(intout, style);
	putArray(ctrl, nINTOUT, 1);
	text_style = style;
}


static void SetFillStyle(void)
{
	word style = getWord(intin);
	if (style > 3) style = 2;
	PutWord(intout, style);
	putArray(ctrl, nINTOUT, 1);
	fill_style = style;
}



static void SetFillIndex(void)
{
	int maxb = (halftone_height > hatches_height ? halftone_height : hatches_height);
	word style = getWord(intin);
	if (style > (maxb / 8 + 1) || style < 1) style = 1;
	PutWord(intout, style);
	putArray(ctrl, nINTOUT, 1);
	fill_index = style;
}


static void InquireColour(void)
{
	int r,g,b;
	int ncol  = getWord(intin);
	int style = getArray(intin, 1);
	if (ncol > 127) ncol = 127;

	putArray(ctrl, nINTOUT, 4);
	PutWord(intout, ncol);

	if (!style)
	{
		putArray(intout, 1, request_col[ncol][0]);
		putArray(intout, 2, request_col[ncol][1]);
		putArray(intout, 3, request_col[ncol][2]);
	}
	else
	{
		GrQueryColor(m_colourMap[ncol], &r, &g, &b);
		putArray(intout, 1, (r * 1000) / 255);
		putArray(intout, 2, (r * 1000) / 255);
		putArray(intout, 3, (r * 1000) / 255);		
	}	

}


static void InquireCellArray(void)
{
	int x , y, rx, ry, c;
	short c1;
	int rowlen  = getArray(ctrl, 5);
	int nrows   = getArray(ctrl, 6);
	int rowused = 0;         
	int nrused  = 0;
	int error   = 0;
	
	getptsin();
	ry = 0;
	for (y = Y0; y < Y1; y++)
	{
		rx  = 0;
		for (x = X0; x < X1; x++)
		{
			c = GrPixel(x, y);
			for (c1 = 0; c1 < 128; c1++)
				if (m_colourMap[c1] == c) break;
			if (c1 == 128) 
			{	
				c1 = (-1);
				error = 1;
			}
			putArray(intin, ry * rowlen + rx, c1);
			++rx; if (rx >= nrused) nrused = rx;
		}
		++ry; if (ry >= nrused) nrused = ry;
		if (nrused > nrows) break;
	}
	putArray(ctrl, 7, rowused);
	putArray(ctrl, 8, nrused);
	putArray(ctrl, 9, error);
}


void PcwGSX::setWriteMode(void)
{
	int mode = getArray(intout, 0);

	if (mode < 1 || mode > 4) mode = 1;

	putArray(ctrl, nINTOUT, 1);
	putArray(intout, 0, mode);

	switch(mode)
	{
		case 1: gsxbg     = m_sysVideo->colourToIndex(0,0,0);
			gsxfg_and = (~0);
			gsxfg_or  = 0;
			break;

		case 2: gsxbg     = 0x00000000;	// Alpha=0 //
			gsxfg_and = (~0);
			gsxfg_or  = 0;
			break;

		case 3: gsxbg	  = 0x00000000; // Transparent //
			gsxfg_and = (~0);
			gsxfg_or  = GrXOR;		// Write in XOR mode //
			break;

		case 4: gsxbg     = GrNOCOLOR; // Transparent //
			gsxfg_and = 0;
			gsxfg_or  = GrBlack();
			break;	
	}
}


static void ReadLocator(void)
{
	GrMouseEvent evt;

	if (!GrMouseDetect()) return;
	if (locator_mode == 1)
	{
		getptsin();
		GrMouseSetCursor(GSXcursor);
		GrMouseWarp(X0, Y0);
		GrMouseDisplayCursor();
		evt.flags = 0;
		GrMouseGetEvent(GR_M_BUTTON_DOWN, &evt);
		
		putArray(ctrl, nINTOUT, 1);
		putArray(ctrl, nPTSOUT, 1);
		putArray(ptsout, 0, (word)evt.x);
		putArray(ptsout, 1, (word)evt.y);
		PutWord(intout,  ((evt.buttons & GR_M_LEFT) ? 0x20 : 0)
				|((evt.buttons & GR_M_MIDDLE) ? 0x21 : 0)
				|((evt.buttons & GR_M_RIGHT) ? 0x22 : 0));
		GrMouseEraseCursor();
	}
	else
	{
		GrMouseSetCursor(GSXcursor);
		GrMouseDisplayCursor();
		evt.flags = 0;
		GrMouseGetEvent(GR_M_POLL | GR_M_MOTION | GR_M_BUTTON_DOWN, &evt);
		
		if (evt.flags & GR_M_MOTION)
		{
			putArray(ctrl, nPTSOUT, 1);
			putArray(ptsout, 0, (word)evt.x);
			putArray(ptsout, 1, (word)evt.y);
		}
		if (evt.flags & GR_M_BUTTON_DOWN)
		{
			putArray(ctrl, nINTOUT, 1);
			PutWord(intout,  ((evt.buttons & GR_M_LEFT) ? 0x20 : 0)
				|((evt.buttons & GR_M_MIDDLE) ? 0x21 : 0)
				|((evt.buttons & GR_M_RIGHT) ? 0x22 : 0));
		}
		GrMouseEraseCursor();
	}
}


static void SetInputMode(void)
{
	int type = getArray(intout, 0);
	int mode = getArray(intout, 1);

	if (type != 1) return;

	if (mode < 1) mode = 1;
	if (mode > 2) mode = 2;

	locator_mode = mode;
}
*/


void PcwGSX::command(word pb)
{
	word function;

	m_surface = m_sys->m_activeTerminal->getSysVideo()->getSurface();
	gsxpb  = pb;
	ctrl   = getWord(pb);
	intin  = getWord(pb + 2);
	ptsin  = getWord(pb + 4);
	intout = getWord(pb + 6);
	ptsout = getWord(pb + 8);	

	putArray(ctrl, nINTOUT, 0);
	putArray(ctrl, nPTSOUT, 0);
//	setGSXopt();
//	setGSXfill();

	function = getWord(ctrl);
	printf("PcwGSX: function=%d\n", function);
	switch(function)
	{
		case 1: openWorkstation();  break;
		case 2: closeWorkstation(); break;
		case 3: clearPicture();	    break;
		case 4:	updateWorkstation();break;
//		case 5: Escape();	    break;
		case 9:	// temporary!
		case 6: drawPolyline();	    break;
//		case 7: DrawPolymarker();   break;
//		case 8: DrawText();	    break;
//		case 9: drawPolygon();	    break;
//		case 10: OutputIndexArray();break;
//		case 11: GDP();		    break;
//		case 12: SetTextHeight;     break;
//		case 13: SetTextDirection();break;
		case 14: setColourIndex();  break;
		case 15: setLineStyle();    break;
//		case 16: SetLineWidth;	    break;
		case 17: setLineColour;     break;
//		case 18: SetMarkerStyle();  break;
//		case 19: SetMarkerHeight;   break;
		case 20: setMarkerColour;   break;
//		case 21: SetTextStyle();    break;
		case 22: setTextColour;     break;
//		case 23: SetFillStyle();    break;
//		case 24: SetFillIndex();    break;
		case 25: setFillColour;     break;
//		case 26: InquireColour();   break;
//		case 27: InquireCellArray();break;
//		case 28: ReadLocator();	    break;
//		case 32: SetWriteMode();    break;
//		case 33: SetInputMode();    break;
		default:		    break;		
	}
}

word PcwGSX::getWord(word addr)
{
	return m_sys->m_cpu->fetchW(addr);
}

word PcwGSX::getArray(word addr, word idx)
{
	return m_sys->m_cpu->fetchW(addr + 2 * idx);
}

void PcwGSX::putArray(word addr, word idx, word value)
{
	m_sys->m_cpu->storeW(addr + 2 * idx, value);
}

bool PcwGSX::parseNode(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
	return true;
}

bool PcwGSX::storeNode(xmlDocPtr doc, xmlNsPtr ns, xmlNodePtr cur)
{
	return true;
}

PcwTerminal& PcwGSX::operator<<(unsigned char c)
{
	return *this;
}


void PcwGSX::setForeground(byte r, byte g, byte b)
{
	if (m_term) m_term->setForeground(r,g,b);
}

void PcwGSX::setBackground(byte r, byte g, byte b)
{
	if (m_term) m_term->setBackground(r,g,b);
}

				        

#endif /* 0 */

